#pragma once

#include <unordered_set>
#include <vulkan/vulkan.hpp>

#include "../config/config.hpp"
#include "queue_structures.hpp"
#include "types.hpp"

namespace core {

  /**
   * @brief Gets the maximum usable sample count for MSAA.
   *
   * @param device Physical device to check
   * @return vk::SampleCountFlagBits Maximum sample count supported
   */
  static fn getMaxUsableSampleCount(const vk::PhysicalDevice& device) -> vk::SampleCountFlagBits {
    vk::PhysicalDeviceProperties physicalDeviceProperties = device.getProperties();

    vk::SampleCountFlags counts = physicalDeviceProperties.limits.framebufferColorSampleCounts &
                                  physicalDeviceProperties.limits.framebufferDepthSampleCounts;

    if (counts & vk::SampleCountFlagBits::e64)
      return vk::SampleCountFlagBits::e64;
    if (counts & vk::SampleCountFlagBits::e32)
      return vk::SampleCountFlagBits::e32;
    if (counts & vk::SampleCountFlagBits::e16)
      return vk::SampleCountFlagBits::e16;
    if (counts & vk::SampleCountFlagBits::e8)
      return vk::SampleCountFlagBits::e8;
    if (counts & vk::SampleCountFlagBits::e4)
      return vk::SampleCountFlagBits::e4;
    if (counts & vk::SampleCountFlagBits::e2)
      return vk::SampleCountFlagBits::e2;

    return vk::SampleCountFlagBits::e1;
  }

  /**
   * @brief Checks if a device supports the required extensions.
   *
   * @param device Physical device to check
   * @return bool True if all required extensions are supported
   */
  static fn checkDeviceExtensionSupport(const vk::PhysicalDevice& device) -> bool {
    // Get the available extensions
    std::vector<vk::ExtensionProperties> availableExtensions = device.enumerateDeviceExtensionProperties();

    // Create a set of required extensions
    std::unordered_set<std::string> requiredExtensions(
      config::deviceExtensions.begin(), config::deviceExtensions.end()
    );

    // For each available extension,
    for (const auto& extension : availableExtensions)
      // Remove it from the required extensions set if it's required
      requiredExtensions.erase(extension.extensionName);

    // If the set is empty, all required extensions are supported
    return requiredExtensions.empty();
  }

  /**
   * @brief Checks if a physical device is suitable for the application.
   *
   * @param device Physical device to check
   * @param surface Surface to check presentation support against
   * @return bool True if device is suitable, false otherwise
   */
  static fn isDeviceSuitable(const vk::PhysicalDevice& device, const vk::SurfaceKHR& surface) -> bool {
    // Get the queue families that support the required operations
    QueueFamilyIndices qfIndices = QueueFamilyIndices::findQueueFamilies(device, surface);

    // Check if the device supports the required extensions
    bool extensionsSupported = checkDeviceExtensionSupport(device);

    bool swapChainAdequate = false;

    if (extensionsSupported) {
      SwapChainSupportDetails swapChainSupport =
        SwapChainSupportDetails::querySwapChainSupport(device, surface);
      // Check if the swap chain is adequate (make sure it has
      // at least one supported format and presentation mode)
      swapChainAdequate = !swapChainSupport.formats.empty() && !swapChainSupport.present_modes.empty();
    }

    // Check if the device supports the required features
    vk::PhysicalDeviceFeatures supportedFeatures = device.getFeatures();

    return qfIndices.isComplete() && extensionsSupported && swapChainAdequate &&
           supportedFeatures.samplerAnisotropy;
  }

} // namespace core
