#pragma once

#include <optional>
#include <vector>
#include <vulkan/vulkan.hpp>

#include "types.hpp"

/**
 * @brief Struct to store queue family indices.
 *
 * This struct contains the indices of the graphics and presentation queue families.
 */
struct QueueFamilyIndices {
  std::optional<u32> graphics_family; ///< Index of graphics queue family
  std::optional<u32> present_family;  ///< Index of presentation queue family

  /**
   * @brief Check if all required queue families are found.
   *
   * @return True if both graphics and presentation families are found, false otherwise.
   */
  fn isComplete() -> bool { return graphics_family.has_value() && present_family.has_value(); }

  /**
   * @brief Finds queue families that support graphics and presentation.
   *
   * @param device Physical device to check
   * @param surface Surface to check presentation support against
   * @return QueueFamilyIndices Struct containing queue family indices
   */
  static fn findQueueFamilies(const vk::PhysicalDevice& device, const vk::SurfaceKHR& surface)
    -> QueueFamilyIndices {
    QueueFamilyIndices indices;

    std::vector<vk::QueueFamilyProperties> queueFamilies = device.getQueueFamilyProperties();

    for (u32 i = 0; i < queueFamilies.size(); i++) {
      const auto& queueFamily = queueFamilies[i];

      // Check for graphics support
      if (queueFamily.queueFlags & vk::QueueFlagBits::eGraphics)
        indices.graphics_family = i;

      // Check for presentation support
      if (device.getSurfaceSupportKHR(i, surface))
        indices.present_family = i;

      if (indices.isComplete())
        break;
    }

    return indices;
  }
};

/**
 * @brief Struct to hold swap chain support details.
 *
 * This struct contains information about the surface capabilities,
 * supported formats, and presentation modes.
 */
struct SwapChainSupportDetails {
  vk::SurfaceCapabilitiesKHR        capabilities;  ///< Surface capabilities
  std::vector<vk::SurfaceFormatKHR> formats;       ///< Supported surface formats
  std::vector<vk::PresentModeKHR>   present_modes; ///< Supported presentation modes

  /**
   * @brief Queries swap chain support details from a physical device.
   *
   * @param device Physical device to query
   * @param surface Surface to check against
   * @return SwapChainSupportDetails Struct containing surface capabilities and supported formats
   */
  static fn querySwapChainSupport(const vk::PhysicalDevice& device, const vk::SurfaceKHR& surface)
    -> SwapChainSupportDetails {
    SwapChainSupportDetails details;

    details.capabilities  = device.getSurfaceCapabilitiesKHR(surface);
    details.formats       = device.getSurfaceFormatsKHR(surface);
    details.present_modes = device.getSurfacePresentModesKHR(surface);

    return details;
  }

  /**
   * @brief Chooses the best surface format for the swap chain.
   *
   * @param availableFormats List of available surface formats
   * @return vk::SurfaceFormatKHR The chosen surface format
   */
  static fn chooseSwapSurfaceFormat(const std::vector<vk::SurfaceFormatKHR>& availableFormats
  ) -> vk::SurfaceFormatKHR {
    // Prefer SRGB with nonlinear color space
    for (const auto& availableFormat : availableFormats)
      if (availableFormat.format == vk::Format::eB8G8R8A8Srgb &&
          availableFormat.colorSpace == vk::ColorSpaceKHR::eSrgbNonlinear)
        return availableFormat;

    // If preferred format not found, use first available
    return availableFormats[0];
  }

  /**
   * @brief Chooses the best presentation mode for the swap chain.
   *
   * @param availablePresentModes List of available presentation modes
   * @return vk::PresentModeKHR The chosen presentation mode
   */
  static fn chooseSwapPresentMode(const std::vector<vk::PresentModeKHR>& availablePresentModes
  ) -> vk::PresentModeKHR {
    // Prefer mailbox mode (triple buffering)
    for (const auto& availablePresentMode : availablePresentModes)
      if (availablePresentMode == vk::PresentModeKHR::eMailbox)
        return availablePresentMode;

    // Fallback to FIFO (vsync)
    return vk::PresentModeKHR::eFifo;
  }

  /**
   * @brief Chooses the swap extent (resolution) for the swap chain.
   *
   * @param capabilities Surface capabilities
   * @param width Desired width
   * @param height Desired height
   * @return vk::Extent2D The chosen swap extent
   */
  static fn chooseSwapExtent(const vk::SurfaceCapabilitiesKHR& capabilities, u32 width, u32 height)
    -> vk::Extent2D {
    if (capabilities.currentExtent.width != std::numeric_limits<u32>::max())
      return capabilities.currentExtent;

    vk::Extent2D actualExtent = { width, height };

    actualExtent.width =
      std::clamp(actualExtent.width, capabilities.minImageExtent.width, capabilities.maxImageExtent.width);

    actualExtent.height =
      std::clamp(actualExtent.height, capabilities.minImageExtent.height, capabilities.maxImageExtent.height);

    return actualExtent;
  }
};
