#pragma once

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include "../core/types.hpp"
#include "vkfw.hpp"

// Camera speed constant from main.cpp
constexpr f64 CAMERA_SPEED = 1.0;

/**
 * @brief Represents a 3D camera in the scene.
 */
struct Camera {
  glm::dvec3 position; ///< Camera's position in 3D space
  glm::dvec3 front;    ///< Direction the camera is facing
  glm::dvec3 up;       ///< Camera's up vector
  glm::dvec3 right;    ///< Camera's right vector
  f64        yaw;      ///< Yaw angle (rotation around vertical axis)
  f64        pitch;    ///< Pitch angle (rotation around horizontal axis)

  /**
   * @brief Constructs a Camera with default settings.
   */
  Camera()
    : position(2.0, 2.0, 2.0),
      front(glm::normalize(glm::dvec3(-2.0, -2.0, -2.0))),
      up(0.0, 0.0, 1.0),
      right(glm::normalize(glm::cross(front, up))),
      yaw(-135.0),
      pitch(-35.26) {
    updateCameraVectors();
  }

  /**
   * @brief Gets the camera's current position.
   * @return The camera's position as a 3D vector.
   */
  [[nodiscard]] fn getPosition() const -> glm::dvec3 { return position; }

  /**
   * @brief Calculates and returns the view matrix for the camera.
   * @return The view matrix as a 4x4 matrix.
   */
  [[nodiscard]] fn getViewMatrix() const -> glm::mat4 { return glm::lookAt(position, position + front, up); }

  /**
   * @brief Moves the camera forward.
   * @param deltaTime Time elapsed since last frame.
   */
  fn moveForward(f64 deltaTime) -> void {
    glm::dvec3 horizontalFront = glm::normalize(glm::dvec3(front.x, front.y, 0.0));
    position += horizontalFront * CAMERA_SPEED * deltaTime;
  }

  /**
   * @brief Moves the camera backward.
   * @param deltaTime Time elapsed since last frame.
   */
  fn moveBackward(f64 deltaTime) -> void {
    glm::dvec3 horizontalFront = glm::normalize(glm::dvec3(front.x, front.y, 0.0));
    position -= horizontalFront * CAMERA_SPEED * deltaTime;
  }

  /**
   * @brief Moves the camera to the left.
   * @param deltaTime Time elapsed since last frame.
   */
  fn moveLeft(f64 deltaTime) -> void {
    glm::dvec3 horizontalRight = glm::normalize(glm::dvec3(right.x, right.y, 0.0));
    position -= horizontalRight * CAMERA_SPEED * deltaTime;
  }

  /**
   * @brief Moves the camera to the right.
   * @param deltaTime Time elapsed since last frame.
   */
  fn moveRight(f64 deltaTime) -> void {
    glm::dvec3 horizontalRight = glm::normalize(glm::dvec3(right.x, right.y, 0.0));
    position += horizontalRight * CAMERA_SPEED * deltaTime;
  }

  /**
   * @brief Moves the camera upward.
   * @param deltaTime Time elapsed since last frame.
   */
  fn moveUp(f64 deltaTime) -> void { position += glm::dvec3(0.0, 0.0, 1.0) * CAMERA_SPEED * deltaTime; }

  /**
   * @brief Moves the camera downward.
   * @param deltaTime Time elapsed since last frame.
   */
  fn moveDown(f64 deltaTime) -> void { position -= glm::dvec3(0.0, 0.0, 1.0) * CAMERA_SPEED * deltaTime; }

  /**
   * @brief Rotates the camera based on mouse movement.
   * @param xoffset Horizontal mouse movement.
   * @param yoffset Vertical mouse movement.
   */
  fn rotate(f64 xoffset, f64 yoffset) -> void {
    const f64 sensitivity = 0.1;
    yaw += xoffset * sensitivity;
    pitch += yoffset * sensitivity;

    pitch = glm::clamp(pitch, -89.0, 89.0);

    updateCameraVectors();
  }

  /**
   * @brief Processes input for camera movement and rotation.
   * @param window The GLFW window.
   * @param camera The camera to be controlled.
   * @param deltaTime Time elapsed since last frame.
   * @param cameraSpeed Speed multiplier for camera movement.
   */
  static fn processInput(vkfw::Window& window, Camera& camera, const f32& deltaTime, const f32& cameraSpeed)
    -> void {
    if (window.getKey(vkfw::Key::eW) == vkfw::eTrue)
      camera.moveForward(static_cast<f64>(deltaTime * cameraSpeed));
    if (window.getKey(vkfw::Key::eA) == vkfw::eTrue)
      camera.moveLeft(static_cast<f64>(deltaTime * cameraSpeed));
    if (window.getKey(vkfw::Key::eS) == vkfw::eTrue)
      camera.moveBackward(static_cast<f64>(deltaTime * cameraSpeed));
    if (window.getKey(vkfw::Key::eD) == vkfw::eTrue)
      camera.moveRight(static_cast<f64>(deltaTime * cameraSpeed));
    if (window.getKey(vkfw::Key::eSpace) == vkfw::eTrue)
      camera.moveUp(static_cast<f64>(deltaTime * cameraSpeed));
    if (window.getKey(vkfw::Key::eLeftShift) == vkfw::eTrue)
      camera.moveDown(static_cast<f64>(deltaTime * cameraSpeed));
  }

 private:
  /**
   * @brief Updates the camera's orientation vectors based on yaw and pitch.
   */
  fn updateCameraVectors() -> void {
    front = glm::normalize(glm::dvec3(
      cos(glm::radians(yaw)) * cos(glm::radians(pitch)),
      sin(glm::radians(yaw)) * cos(glm::radians(pitch)),
      sin(glm::radians(pitch))
    ));
    right = glm::normalize(glm::cross(front, glm::dvec3(0.0, 0.0, 1.0)));
    up    = glm::normalize(glm::cross(right, front));
  }
};
