#pragma once

#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#include "../core/types.hpp"

namespace graphics {

  /**
   * @brief Creates a Vulkan image view.
   *
   * This function creates and returns a unique Vulkan image view using the provided parameters.
   *
   * @param device The logical device to create the view on
   * @param image The Vulkan image for which to create the view
   * @param format The format of the image
   * @param aspectFlags The aspect flags for the image view
   * @param mipLevels The number of mip levels for the image view
   * @return vk::UniqueImageView A unique handle to the created Vulkan image view
   *
   * @details
   * The function creates an image view with the following properties:
   * - 2D view type
   * - Subresource range starting from base mip level 0
   * - Single array layer starting from base array layer 0
   */
  static fn createImageView(
    const vk::Device&           device,
    const vk::Image&            image,
    const vk::Format&           format,
    const vk::ImageAspectFlags& aspectFlags,
    const u32&                  mipLevels
  ) -> vk::UniqueImageView {
    return device.createImageViewUnique({
      .image            = image,
      .viewType         = vk::ImageViewType::e2D,
      .format           = format,
      .subresourceRange = {
                           .aspectMask     = aspectFlags,
                           .baseMipLevel   = 0,
                           .levelCount     = mipLevels,
                           .baseArrayLayer = 0,
                           .layerCount     = 1,
                         },
    });
  }

  /**
   * @brief Finds a suitable memory type for allocation.
   *
   * @param physicalDevice The physical device to check
   * @param typeFilter Filter for memory types
   * @param properties Required memory properties
   * @return u32 Index of the suitable memory type
   */
  static fn findMemoryType(
    const vk::PhysicalDevice&      physicalDevice,
    const u32&                     typeFilter,
    const vk::MemoryPropertyFlags& properties
  ) -> u32 {
    vk::PhysicalDeviceMemoryProperties memProperties = physicalDevice.getMemoryProperties();

    for (u32 i = 0; i < memProperties.memoryTypeCount; i++)
      if ((typeFilter & (1 << i)) && (memProperties.memoryTypes[i].propertyFlags & properties) == properties)
        return i;

    throw std::runtime_error("Failed to find suitable memory type!");
  }

  /**
   * @brief Creates a Vulkan image and allocates memory for it.
   *
   * @param device The logical device to create the image on
   * @param width Width of the image
   * @param height Height of the image
   * @param mipLevels Number of mip levels
   * @param numSamples Number of samples for multisampling
   * @param format Format of the image
   * @param tiling Tiling mode of the image
   * @param usage Usage flags for the image
   * @param properties Memory property flags
   * @return std::pair<vk::UniqueImage, vk::UniqueDeviceMemory> A pair containing the image and its memory
   */
  static fn createImage(
    const vk::Device&              device,
    const vk::PhysicalDevice&      physicalDevice,
    const u32&                     width,
    const u32&                     height,
    const u32&                     mipLevels,
    const vk::SampleCountFlagBits& numSamples,
    const vk::Format&              format,
    const vk::ImageTiling&         tiling,
    const vk::ImageUsageFlags&     usage,
    const vk::MemoryPropertyFlags& properties
  ) -> std::pair<vk::UniqueImage, vk::UniqueDeviceMemory> {
    // Define the image creation info
    vk::ImageCreateInfo imageInfo {
      .imageType     = vk::ImageType::e2D,
      .format        = format,
      .extent        = { .width = width, .height = height, .depth = 1 },
      .mipLevels     = mipLevels,
      .arrayLayers   = 1,
      .samples       = numSamples,
      .tiling        = tiling,
      .usage         = usage,
      .sharingMode   = vk::SharingMode::eExclusive,
      .initialLayout = vk::ImageLayout::eUndefined,
    };

    // Create the image
    vk::UniqueImage image = device.createImageUnique(imageInfo);

    // Get the memory requirements for the image
    vk::MemoryRequirements memRequirements = device.getImageMemoryRequirements(image.get());

    // Memory allocation info
    vk::MemoryAllocateInfo allocInfo {
      .allocationSize  = memRequirements.size,
      .memoryTypeIndex = findMemoryType(physicalDevice, memRequirements.memoryTypeBits, properties),
    };

    // Allocate memory
    vk::UniqueDeviceMemory imageMemory = device.allocateMemoryUnique(allocInfo);

    // Bind the allocated memory to the image
    device.bindImageMemory(image.get(), imageMemory.get(), 0);

    return { std::move(image), std::move(imageMemory) };
  }

  /**
   * @brief Transitions an image between layouts.
   *
   * @param device The logical device
   * @param commandPool The command pool to allocate command buffers from
   * @param queue The queue to submit commands to
   * @param image The image to transition
   * @param oldLayout The old layout
   * @param newLayout The new layout
   * @param mipLevels Number of mip levels
   */
  static fn transitionImageLayout(
    const vk::Device&      device,
    const vk::CommandPool& commandPool,
    const vk::Queue&       queue,
    const vk::Image&       image,
    const vk::ImageLayout& oldLayout,
    const vk::ImageLayout& newLayout,
    const u32&             mipLevels
  ) -> void {
    // Create a command buffer
    vk::CommandBufferAllocateInfo allocInfo {
      .commandPool        = commandPool,
      .level              = vk::CommandBufferLevel::ePrimary,
      .commandBufferCount = 1,
    };

    vk::UniqueCommandBuffer commandBuffer = std::move(device.allocateCommandBuffersUnique(allocInfo)[0]);

    // Begin recording
    commandBuffer->begin({ .flags = vk::CommandBufferUsageFlagBits::eOneTimeSubmit });

    // Define the image memory barrier
    vk::ImageMemoryBarrier barrier {
      .oldLayout           = oldLayout,
      .newLayout           = newLayout,
      .srcQueueFamilyIndex = vk::QueueFamilyIgnored,
      .dstQueueFamilyIndex = vk::QueueFamilyIgnored,
      .image               = image,
      .subresourceRange    = {
        .aspectMask     = vk::ImageAspectFlagBits::eColor,
        .baseMipLevel   = 0,
        .levelCount     = mipLevels,
        .baseArrayLayer = 0,
        .layerCount     = 1,
      },
    };

    // Define the source and destination stages
    vk::PipelineStageFlags sourceStage;
    vk::PipelineStageFlags destinationStage;

    // Define the access masks
    if (oldLayout == vk::ImageLayout::eUndefined && newLayout == vk::ImageLayout::eTransferDstOptimal) {
      barrier.srcAccessMask = {};
      barrier.dstAccessMask = vk::AccessFlagBits::eTransferWrite;

      sourceStage      = vk::PipelineStageFlagBits::eTopOfPipe;
      destinationStage = vk::PipelineStageFlagBits::eTransfer;
    } else if (oldLayout == vk::ImageLayout::eTransferDstOptimal &&
               newLayout == vk::ImageLayout::eShaderReadOnlyOptimal) {
      barrier.srcAccessMask = vk::AccessFlagBits::eTransferWrite;
      barrier.dstAccessMask = vk::AccessFlagBits::eShaderRead;

      sourceStage      = vk::PipelineStageFlagBits::eTransfer;
      destinationStage = vk::PipelineStageFlagBits::eFragmentShader;
    } else {
      // Ensure that the layout transition is supported
      throw std::invalid_argument("Unsupported layout transition!");
    }

    // Record the pipeline barrier
    commandBuffer->pipelineBarrier(sourceStage, destinationStage, {}, {}, {}, barrier);

    // End recording
    commandBuffer->end();

    // Submit the command buffer
    vk::SubmitInfo submitInfo {
      .commandBufferCount = 1,
      .pCommandBuffers    = &commandBuffer.get(),
    };

    queue.submit(submitInfo);
    queue.waitIdle();
  }

  /**
   * @brief Copies data from a buffer to an image.
   *
   * @param device The logical device
   * @param commandPool The command pool to allocate command buffers from
   * @param queue The queue to submit commands to
   * @param buffer Source buffer
   * @param image Destination image
   * @param width Image width
   * @param height Image height
   */
  static fn copyBufferToImage(
    const vk::Device&      device,
    const vk::CommandPool& commandPool,
    const vk::Queue&       queue,
    const vk::Buffer&      buffer,
    const vk::Image&       image,
    const u32&             width,
    const u32&             height
  ) -> void {
    // Create a command buffer
    vk::CommandBufferAllocateInfo allocInfo {
      .commandPool        = commandPool,
      .level              = vk::CommandBufferLevel::ePrimary,
      .commandBufferCount = 1,
    };

    vk::UniqueCommandBuffer commandBuffer = std::move(device.allocateCommandBuffersUnique(allocInfo)[0]);

    // Begin recording
    commandBuffer->begin({ .flags = vk::CommandBufferUsageFlagBits::eOneTimeSubmit });

    vk::BufferImageCopy region {
      .bufferOffset      = 0,
      .bufferRowLength   = 0,
      .bufferImageHeight = 0,
      .imageSubresource  = { .aspectMask     = vk::ImageAspectFlagBits::eColor,
                            .mipLevel       = 0,
                            .baseArrayLayer = 0,
                            .layerCount     = 1 },
      .imageOffset       = { .x = 0, .y = 0, .z = 0 },
      .imageExtent       = { .width = width, .height = height, .depth = 1 },
    };

    commandBuffer->copyBufferToImage(buffer, image, vk::ImageLayout::eTransferDstOptimal, 1, &region);

    // End recording
    commandBuffer->end();

    // Submit the command buffer
    vk::SubmitInfo submitInfo {
      .commandBufferCount = 1,
      .pCommandBuffers    = &commandBuffer.get(),
    };

    queue.submit(submitInfo);
    queue.waitIdle();
  }

} // namespace graphics
