/**
 * @file vertex.hpp
 * @brief Defines the vertex structure and its associated utilities for 3D rendering.
 *
 * This file contains the Vertex structure used for 3D model representation in the Vulkan
 * graphics pipeline. It includes position, color, and texture coordinate data, along with
 * Vulkan-specific descriptors for vertex input handling.
 */

#define GLM_FORCE_DEPTH_ZERO_TO_ONE
#define GLM_ENABLE_EXPERIMENTAL
#include <glm/glm.hpp>
#include <glm/gtx/hash.hpp>

#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#include "../core/types.hpp"

/**
 * @brief Represents a vertex in 3D space with color and texture information.
 *
 * This structure defines a vertex with all its attributes required for rendering,
 * including position in 3D space, RGB color, and texture coordinates. It also
 * provides methods for Vulkan vertex input configuration.
 */
struct Vertex {
  vec3 pos;       ///< Position of the vertex in 3D space (x, y, z)
  vec3 color;     ///< RGB color values, each component in range [0, 1]
  vec2 tex_coord; ///< Texture coordinates (u, v) for texture mapping

  /**
   * @brief Provides the vertex binding description for Vulkan.
   *
   * @return vk::VertexInputBindingDescription Describes how to bind vertex data to GPU memory.
   *
   * The binding description specifies:
   * - Binding index (0)
   * - Stride (size of one vertex)
   * - Input rate (per-vertex data)
   */
  static fn getBindingDescription() -> vk::VertexInputBindingDescription {
    return { .binding = 0, .stride = sizeof(Vertex), .inputRate = vk::VertexInputRate::eVertex };
  }

  /**
   * @brief Provides attribute descriptions for vertex data interpretation.
   *
   * @return std::array<vk::VertexInputAttributeDescription, 3> Array of descriptions for position, color, and
   * texture coordinates.
   *
   * The attribute descriptions specify:
   * - Location indices (0 for position, 1 for color, 2 for texture coordinates)
   * - Binding point (0)
   * - Data format (R32G32B32 for vec3, R32G32 for vec2)
   * - Offset of each attribute in the vertex structure
   */
  static fn getAttributeDescriptions() -> std::array<vk::VertexInputAttributeDescription, 3> {
    return {
      vk::VertexInputAttributeDescription { 0, 0, vk::Format::eR32G32B32Sfloat, offsetof(Vertex,       pos) },
      vk::VertexInputAttributeDescription { 1, 0, vk::Format::eR32G32B32Sfloat, offsetof(Vertex,     color) },
      vk::VertexInputAttributeDescription { 2, 0,    vk::Format::eR32G32Sfloat, offsetof(Vertex, tex_coord) }
    };
  }

  /**
   * @brief Compares two vertices for equality.
   *
   * @param other The vertex to compare with.
   * @return bool True if vertices are identical in position, color, and texture coordinates.
   */
  fn operator==(const Vertex& other) const->bool {
    return pos == other.pos && color == other.color && tex_coord == other.tex_coord;
  }
};

namespace std {
  /**
   * @brief Hash function specialization for Vertex type.
   *
   * This specialization allows Vertex objects to be used as keys in unordered containers.
   * The hash combines position, color, and texture coordinate data using bit operations
   * to create a unique hash value.
   */
  template <>
  struct hash<Vertex> {
    /**
     * @brief Computes hash value for a vertex.
     *
     * @param vertex The vertex to hash.
     * @return size_t Hash value combining all vertex attributes.
     */
    fn operator()(Vertex const& vertex) const->size_t {
      return ((hash<vec3>()(vertex.pos) ^ (hash<vec3>()(vertex.color) << 1)) >> 1) ^
             (hash<vec2>()(vertex.tex_coord) << 1);
    }
  };
}
